import pandas as pd
import requests
import sys
from datapungi_fed import generalSettings 
from datapungi_fed import drivers
from datapungi_fed.driverCore import driverCore
#from driverCore import driverCore
#import drivers

#TODO: test clipcode, utils (setting folder, options), getting help in each step.
class data():
    '''
       Connect to all of FRED and GeoFred databases. To start, choose a database group from:

       - categories: 8 top categories (eg, National Accounts) and their subgroups
                   https://fred.stlouisfed.org/categories/
       - releases: about 300 datasets of time series grouped by release (eg, Penn World Table)
                   https://fred.stlouisfed.org/releases/
       - series: time series data
                   https://fred.stlouisfed.org/
       - sources: about 90 data providers (eg, IMF and Bank of Mexico)
                   https://fred.stlouisfed.org/sources/
       - tags: 	Tags applied to time series (eg, location, data source, frequency) - about 5,000 tags
                   https://fred.stlouisfed.org/tags/
       - geo: spatial economic data
                   https://geofred.stlouisfed.org/

       For example:
       import datapungi_fed as dpf

       data = dpf.data()
       print(data.series)

       This class, and datapungi_fed itself, defaults to the series database.  Hence, the following
       will query the series database:

       import datapungi_fed as dpf

       dfp('gdp')
       data = dpf.data()
       data('gdp')

       Check the __doc__ of each database for more information.
    '''
    def __init__(self,connectionParameters = {}, userSettings = {}):
        self.__connectInfo = generalSettings.getGeneralSettings(connectionParameters = connectionParameters, userSettings = userSettings ) 
        self._metadata = self.__connectInfo.packageMetadata
        self._help     = self.__connectInfo.datasourceOverview
        #load drivers:
        loadInfo = {'baseRequest' : self.__connectInfo.baseRequest, 'connectionParameters' : self.__connectInfo.connectionParameters}
        
        #specific drivers
        self.datasetlist  = drivers.datasetlist(**loadInfo)
        
        #core drivers
        coreDriversParams = driverCore()
        for dbGroupName in [x['group'] for x in coreDriversParams._dbParams]:
            setattr(self, dbGroupName.lower(), driverCore(dbGroupName,**loadInfo))
             
    def __call__(self,*args,**kwargs):
        return(self.series(*args,**kwargs))

    def __str__(self):
        return(self.__doc__)

    def _clipcode(self):
        try:
            self._lastCalledDriver.clipcode()
        except:
            print('Get data using a driver first, eg: ')
            #eg: data.NIPA("T10101", verbose = True)
    

if __name__ == '__main__':            
    d = data()
    #print(d)
    #print(d.datasetlist())   
    #print(d.categories(125))   
    #print(d.releases())   
    #print(d.series('GDP'))
    #print(d('GNP'))
    #print(d.sources('1'))   
    #print(d.tags(tag_names='monetary+aggregates;weekly'))   
    #print(d.geo['shapes']('bea'))
    #print(d.geo['meta']('SMU56000000500000001a'))
    #v= d.geo['series'](series_id='WIPCPI',start_date='2012-01-01',verbose=True)
    #print(v)

    #v= d.geo['data'](series_group='882',date='2013-01-01',region_type='state',units='Dollars',frequency='a',season='NSA')
    #print(v)
    #print(d.geo(series_id='WIPCPI',start_date='2012-01-01'))
    #print(d.geo.__doc__)
    d.categories(125)


'''
   Drivers that do not fit the driverCore format.
'''

import pandas as pd
import requests
import json
from copy import deepcopy
import pyperclip
import math
import re
import inspect
import yaml
import itertools
import warnings

from datetime import datetime
from datapungi_fed import generalSettings  # NOTE: projectName
#import generalSettings        #NOTE: projectName
from datapungi_fed import utils  # NOTE: projectName
#import utils                  #NOTE: projectName
from datapungi_fed.driverCore import driverCore
#from driverCore import driverCore


class datasetlist(driverCore):
    def _query(self):
        '''
         Returns name of available datasets, a short description and their query parameters.
         Args:
           none
         Output:
           - pandas table with query function name, database name, short description and query parameters.
        '''
        #get all dictionary of all drivers (in config/datasetlist.yaml)
        datasetlist = self._dbParams
        datasetlistExp = [[{**entry, **dataset}
                           for dataset in entry.pop('datasets')] for entry in datasetlist]
        datasetlistFlat = list(itertools.chain.from_iterable(
            datasetlistExp))  # flatten the array of array
        df_output = pd.DataFrame(datasetlistFlat)
        return(df_output)
    def __call__(self):
        return(self._query())
    

if __name__ == '__main__':
    d = datasetlist()
    v = d(); print(v)


import datapungi_fed as dp 
import time
import pandas as pd
import os


def executeCode(stringIn):
    '''
      auxiliary function for tests: get the requests code as a string and try to execute it.
    '''
    try:
        exec(stringIn+'\n')      #exec('print("hi")') #
        return(dict( codeRun = True, codeOutput = locals()['df_output']   ))   #try to output the dataframe called df_output
    except:
        try:
            exec(stringIn)  #if no dataframe called output, try to see it at least can exec the code
            return(dict(codeRun = True, codeOutput = pd.DataFrame([])))
        except:
            return(dict(codeRun = False, codeOutput = pd.DataFrame([])))

# start  the driver - used by all tests
def startDriver(cmdopt):
    if not cmdopt == "":
        connectionParameters = {"key": cmdopt, "url": ""}
    else:
        connectionParameters = {}
    data = dp.data(connectionParameters)
    return(data)    

# content of test_sample.py
def test_startDriver(cmdopt):
    data = startDriver(cmdopt)
    assert data


###########################################################################################################
###  Tests of the categories dbgroup

def test_categories(cmdopt):
    data = startDriver(cmdopt)
    driver = data.categories(125,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  

def test_categoriesLong(cmdopt):
    data = startDriver(cmdopt)
    driver = data.categories['category'](125,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  

def test_categoriesChildren(cmdopt):
    data = startDriver(cmdopt)
    driver = data.categories['children'](13,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  

def test_categoriesRelated(cmdopt):
    data = startDriver(cmdopt)
    driver = data.categories['related'](32073,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  

def test_categoriesSeries(cmdopt):
    data = startDriver(cmdopt)
    driver = data.categories['series'](125,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  

def test_categoriesTags(cmdopt):
    data = startDriver(cmdopt)
    driver = data.categories['tags'](125,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  

def test_categoriesRelatedTags(cmdopt):
    data = startDriver(cmdopt)
    driver = data.categories['related_tags'](125,tag_names="services;quarterly",verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  

###########################################################################################################
###  Tests of the releases dbgroup

def test_releases(cmdopt):
    data = startDriver(cmdopt)
    driver = data.releases(verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  

def test_releasesLong(cmdopt):
    data = startDriver(cmdopt)
    driver = data.releases['releases'](verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  

def test_releasesDates(cmdopt):
    data = startDriver(cmdopt)
    driver = data.releases['release/dates'](release_id=53,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  

def test_releasesRelease(cmdopt):
    data = startDriver(cmdopt)
    driver = data.releases['release'](release_id=53,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_releasesReleaseDates(cmdopt):
    data = startDriver(cmdopt)
    driver = data.releases['release/dates'](release_id=53,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_releasesReleaseSeries(cmdopt):
    data = startDriver(cmdopt)
    driver = data.releases['release/series'](release_id=53,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_releasesReleaseSources(cmdopt):
    data = startDriver(cmdopt)
    driver = data.releases['release/sources'](release_id=53,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_releasesReleaseTags(cmdopt):
    data = startDriver(cmdopt)
    driver = data.releases['release/tags'](release_id=53,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_releasesReleaseRelatedTags(cmdopt):
    data = startDriver(cmdopt)
    driver = data.releases['release/related_tags'](release_id='86',tag_names='sa;foreign',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_releasesReleaseTables(cmdopt):
    data = startDriver(cmdopt)
    driver = data.releases['release/tables'](release_id=53,verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

###########################################################################################################
## Test Series

def test_series(cmdopt):
    data = startDriver(cmdopt)
    driver = data.series('gdp',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  

def test_seriesSeries(cmdopt):
    data = startDriver(cmdopt)
    driver = data.series['series']('GDP',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  

def test_seriesCategories(cmdopt):
    data = startDriver(cmdopt)
    driver = data.series['categories']('EXJPUS',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_seriesObservations(cmdopt): #seriesLong
    data = startDriver(cmdopt)
    driver = data.series['observations']('GNP',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_seriesRelease(cmdopt):
    data = startDriver(cmdopt)
    driver = data.series['release']('IRA',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_seriesSearch(cmdopt):
    data = startDriver(cmdopt)
    driver = data.series['search'](search_text='monetary+service+index',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_seriesSearchTags(cmdopt):
    data = startDriver(cmdopt)
    driver = data.series['search/tags'](series_search_text='monetary+service+index',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_seriesSearchRelatedTags(cmdopt):
    data = startDriver(cmdopt)
    driver = data.series['search/related_tags'](series_search_text='mortgage+rate',tag_names='30-year;frb',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_seriesTags(cmdopt):
    data = startDriver(cmdopt)
    driver = data.series['tags'](series_id='STLFSI',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_seriesCategories(cmdopt):
    data = startDriver(cmdopt)
    driver = data.series['categories']('EXJPUS',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_seriesUpdates(cmdopt):
    data = startDriver(cmdopt)
    driver = data.series['updates'](verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

def test_seriesVintagedates(cmdopt):
    data = startDriver(cmdopt)
    driver = data.series['vintagedates']('GNPCA',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the 

###########################################################################################################
## Test Tags

def test_tags(cmdopt):
    data = startDriver(cmdopt)
    driver = data.tags(tag_names='monetary+aggregates;weekly',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    assert execCode['codeOutput'].equals(driver['dataFrame']) #test if the output of the code equals the output of the  
 
def test_tagsLong(cmdopt):   
    data = startDriver(cmdopt)
    driver = data.tags['tags'](tag_names='monetary+aggregates;weekly',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.

def test_tagsRelated(cmdopt): 
    data = startDriver(cmdopt)
    driver = data.tags['related_tags'](tag_names='monetary+aggregates;weekly',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    
def test_tagsSeries(cmdopt): 
    data = startDriver(cmdopt)    
    driver = data.tags['tags/series'](tag_names='slovenia;food;oecd',verbose=True)
    execCode = executeCode(driver['code']) 
    assert driver['request'].status_code == 200  #test if connection was stablished
    assert not driver['dataFrame'].empty         #cleaned up output is not empty
    assert execCode['codeRun']                   #try to execute the code.
    


if __name__ == '__main__':
    test_tags()
    #test_
    #test_
    #test_
    #test_
    #test_


'''
   Base driver class
'''

import pandas as pd
import requests
import json
from copy import deepcopy
import pyperclip
import math
import re
import inspect
import yaml
import itertools
from datetime import datetime
import warnings
import functools
from textwrap import dedent
from datapungi_fed import generalSettings        #NOTE: projectName 
#import generalSettings        #NOTE: projectName 
from datapungi_fed import utils                  #NOTE: projectName  
#import utils                  #NOTE: projectName  

class driverCore():
    r'''
      Given a dbGroupName and its default db, starts a factory of query functions - ie, a function for 
      each db in the group.  If dbGroupName is empty, return the list of dbGroups, dbs in the group, and their parameters
    '''
    def __init__(self,dbGroupName='', baseRequest={},connectionParameters={},userSettings={}):  
        #TODO: place defaultQueryFactoryEntry in yaml
        self._dbParams, self.defaultQueryFactoryEntry  = self._getDBParameters(dbGroupName) 
        self._ETDB      = extractTransformDB(baseRequest,connectionParameters,userSettings) #a generic query is started
        self._ETFactory = extractTransformFactory(dbGroupName,self._ETDB,self._dbParams,self.defaultQueryFactoryEntry)
        self._driverMeta = driverMetadata()(dbGroupName)
        self.__setdoc__(dbGroupName)
    
    def __getitem__(self,dbName):
        return(self._ETFactory.extractTransformFactory[dbName])
    
    def __call__(self,*args,**kwargs):
        out = self._ETFactory.extractTransformFactory[self.defaultQueryFactoryEntry](*args,**kwargs)
        return(out)
    
    def __setdoc__(self,dbGroupName):
        if dbGroupName == '':
            self.__doc__ = 'Returns the metadata of the dataset groups and their databases. Do not need inputs.'
        else:
            self.__doc__ = 'Queries the databases of {} \n \n'.format(dbGroupName)
            for entry in self.__docParams__: 
                self.__doc__ += '- {short name}: {description} \n'.format(**entry)
                self.__doc__ += '       parameters: {}\n'.format(str(entry['parameters']))
                self.__doc__ += '       official database name: {}\n'.format(entry['database'])
            self.__doc__ += '\nDefault query database: {}\n'.format(self.defaultQueryFactoryEntry) 
            self.__doc__ += "Sample functions: \n-data.{dbGroupName}() (default)  \n-data.{dbGroupName}['{db}']() (query the {db} database)".format(**{'dbGroupName':dbGroupName.lower(),'db':self.defaultQueryFactoryEntry})       
            self.__doc__ += "\n\nNOTE: don't need to pass most parameters.  Eg, api_key and file_type (json)."
    def __str__(self):
        return(self.__doc__)
           
    def _getDBParameters(self,dbGroupName = ''):
        r'''
          The parameters of each database in the group (if empty returns all groups x databases)
        '''  
        dataPath = utils.getResourcePath('/config/datasetlist.yaml')
        with open(dataPath, 'r') as yf:
            datasetlist = yaml.safe_load(yf)
        
        if dbGroupName == '':
            defaultDB = {}
            return((datasetlist,defaultDB))
        
        #get the entry of the group:
        selected = list(filter( lambda x: x['group'] == dbGroupName , datasetlist))[0]
        defaultDB = selected.get('default query','')
        datasets = selected.get('datasets',{})
        removeCases = lambda array: list(filter( lambda x: x not in ['api_key','file_type']  , array ))
        dbParams = { entry['short name'] : { 'urlSuffix' : entry['database'] , 'json key': entry['json key'], 'params': removeCases(entry['parameters']) } for entry in datasets }
        
        self.__docParams__ =  datasets #parameters used to write a doc string for the class instance.
        return((dbParams,defaultDB))


class extractTransformFactory():
    r'''
      given a groupName of databases, constructs dictionary of functions querying all of its databases
    '''
    def __init__(self,dbGroupName,ETDB,dbParams,defaultQueryFactoryEntry):
        if dbGroupName:
            self.dbGroupName = dbGroupName
            self.dbParams = dbParams
            self.ETDB = ETDB
            self.ETDB(self.dbGroupName,self.dbParams)  #update the connector to the databases with parameters specific to the collection of dbs.
            self.extractTransformFactory = { dbName : self.selectDBQuery(self.query, dbName)  for dbName in self.dbParams.keys() }
            self.defaultQueryFactoryEntry = defaultQueryFactoryEntry  #the entry in query factory that __call__ will use.
        else:
            self.extractTransformFactory = {}
    
    def query(self,*args,**kwargs):
        return( self.ETDB.query(*args,**kwargs) )
    
    def selectDBQuery(self,queryFun,dbName):
        r'''
          Fix a generic query to a query to dbName, creates a lambda that, from
          args/kwargs creates a query of the dbName 
        '''
        fun  = functools.partial(queryFun,dbName)
        lfun = lambda *args,**kwargs: fun(**self.getQueryArgs(dbName,*args,**kwargs))
        #add quick user tips
        lfun.options = self.dbParams[dbName]['params']
        return(lfun)
    
    def getQueryArgs(self,dbName,*args,**kwargs):
        r'''
          Map args and kwargs to driver args
        '''
        #paramaters to be passed to a requests query:
        paramArray = self.dbParams[dbName]['params']
        params = dict(zip(paramArray,args))
        paramsAdd = {key:val for key, val in kwargs.items() if key in paramArray}
        params.update(paramsAdd)
        #non query options (eg, verbose)
        otherArgs = {key:val for key, val in kwargs.items() if not key in paramArray}
        return({**{'params':params},**otherArgs})


class extractTransformDB():
    r'''
      Functions to connect and query a db given its dbName and dbParams (see yaml in config for these).
    '''
    def __init__(self,baseRequest={},connectionParameters={},userSettings={}):
        '''
          loads generic parametes (ie api key, location fo data.)
        '''
        self._connectionInfo = generalSettings.getGeneralSettings(connectionParameters = connectionParameters, userSettings = userSettings )
        self._baseRequest    = self.getBaseRequest(baseRequest,connectionParameters,userSettings)  
        self._lastLoad       = {}  #data stored here to assist functions such as clipcode 
        self._transformData  = transformExtractedData()   
        self._getCode        = transformIncludeCodeSnippet()  
        self._cleanCode      = ""  #TODO: improvable - this is the code snippet producing a pandas df
    
    def __call__(self,dbGroup,dbParams):
        r'''
         A call to an instance of the class Loads specific parameters of the dbs of dbGroup 
        '''
        self.dbGroup  = dbGroup
        self.dbParams = dbParams

    def query(self,dbName,params={},file_type='json',verbose=False,warningsOn=True):
        r'''
          Args:
            params
            file_type              
            verbose             
            warningsOn      
        '''
        # get requests' query inputs
        warningsList = ['countPassLimit']  # warn on this events.
        prefixUrl = self.dbParams[dbName]['urlSuffix']
        output = self.queryApiCleanOutput(prefixUrl, dbName, params, warningsList, warningsOn, verbose)
        return(output) 
     
    def queryApiCleanOutput(self,urlPrefix,dbName,params,warningsList,warningsOn,verbose):
        r'''
            Core steps of querying and cleaning data.  Notice, specific data cleaning should be 
            implemented in the specific driver classes

            Args:
                self - should containg a base request (url)
                urlPrefix (str) - a string to be appended to request url (eg, https:// ...// -> https//...//urlPrefix?)
                
                params (dict) - usually empty, override any query params with the entries of this dictionary
                warningsList (list) - the list of events that can lead to warnings
                warningsOn (bool) - turn on/off driver warnings
                verbose (bool) - detailed output or short output
        '''
        #get data 
        query = self.getBaseQuery(urlPrefix,params)
        retrivedData = requests.get(** { key:entry for key, entry in query.items() if key in ['params','url'] } )
        
        #clean data
        df_output,self._cleanCode = self.cleanOutput(dbName,query,retrivedData)
        
        #print warning if there is more data the limit to download
        for entry in warningsList:
            self._warnings(entry,retrivedData,warningsOn) 
        
        #short or detailed output, update _lastLoad attribute:
        output = self.formatOutputupdateLoadedAttrib(query,df_output,retrivedData,verbose)
        
        return(output)
    
    def getBaseQuery(self,urlPrefix,params):
        r'''
          Return a dictionary of request arguments.

          Args:
              urlPrefix (str) - string appended to the end of the core url (eg, series -> http:...\series? )
              dbName (str) - the name of the db being queried
              params (dict) - a dictionary with request paramters used to override all other given parameters
          Returns:
              query (dict) - a dictionary with 'url' and 'params' (a string) to be passed to a request
        '''
        query = deepcopy(self._baseRequest)
        
        #update query url
        query['url'] = query['url']+urlPrefix   
        query['params'].update(params)
        query['params_dict'] = query['params']
        query['params'] = '&'.join([str(entry[0]) + "=" + str(entry[1]) for entry in query['params'].items()])
        
        return(query)
    
    def formatOutputupdateLoadedAttrib(self,query,df_output,retrivedData,verbose):
        if verbose == False:
            self._lastLoad = df_output
            return(df_output)
        else:
            code = self._getCode.transformIncludeCodeSnippet(query,self._baseRequest,self._connectionInfo.userSettings,self._cleanCode)
            output = dict(dataFrame = df_output, request = retrivedData, code = code)  
            self._lastLoad = output
            return(output)
    
    def cleanOutput(self,dbName,query,retrivedData):
        r'''
         This is a placeholder - specific drivers should have their own cleaning method
         this generates self._cleanCode
        '''
        transformedOutput = self._transformData(self.dbGroup,dbName,self.dbParams,query,retrivedData)
        return(transformedOutput)
    
    def getBaseRequest(self,baseRequest={},connectionParameters={},userSettings={}):
        r'''
          Write a base request.  This is the information that gets used in most requests such as getting the userKey
        '''
        if baseRequest =={}:
           connectInfo = generalSettings.getGeneralSettings(connectionParameters = connectionParameters, userSettings = userSettings )
           return(connectInfo.baseRequest)
        else:
           return(baseRequest)
    
    def _warnings(self,warningName,inputs,warningsOn = True):
        if not warningsOn:
            return
        
        if warningName == 'countPassLimit':
            '''
              warns if number of lines in database exceeds the number that can be downloaded.
              inputs = a request result of a FED API 
            '''
            _count = inputs.json().get('count',1)
            _limit = inputs.json().get('limit',1000)
            if _count > _limit:
              warningText = 'NOTICE: dataset exceeds download limit! Check - count ({}) and limit ({})'.format(_count,_limit)
              warnings.warn(warningText) 


class transformExtractedData():
    def __call__(self,dbGroup,dbName,dbParams,query,retrivedData):
        if dbGroup == 'Series':
            return( self.cleanOutputSeries(dbName,dbParams,query,retrivedData) )
        if dbGroup == 'Geo':
            return( self.cleanOutputGeo(dbName,dbParams,query,retrivedData) )
        else:
            return( self.cleanOutput(dbName,dbParams,query,retrivedData) )
      
    def cleanOutput(self, dbName, dbParams,query, retrivedData): #categories, releases, sources, tags
        dataKey = dbParams[dbName]['json key']
        cleanCode = "df_output =  pd.DataFrame( retrivedData.json()['{}'] )".format(dataKey)
        df_output = pd.DataFrame(retrivedData.json()[dataKey])  # TODO: deal with xml
        warnings.filterwarnings("ignore", category=UserWarning)
        setattr(df_output, '_meta', dict(filter(lambda entry: entry[0] != dataKey, retrivedData.json().items())))  
        warnings.filterwarnings("always", category=UserWarning)
        return((df_output,cleanCode))
    
    def cleanOutputSeries(self, dbName, dbParams,query, retrivedData): #series
        dataKey = dbParams[dbName]['json key']
        cleanCode = "df_output =  pd.DataFrame( retrivedData.json()['{}'] )".format(dataKey)
        df_output = pd.DataFrame(retrivedData.json()[dataKey])  # TODO: deal with xml
        if dbName == 'observations':
            seriesID =  query['params_dict']['series_id'] #{ x.split('=')[0] : x.split('=')[1] for x in query['params'].split("&") }['series_id'] 
            df_output = (df_output[['date','value']]
                .assign( dropRow = lambda df: pd.to_numeric(df['value'],errors='coerce')   )
                .dropna()
                .drop('dropRow',axis=1)
                .assign(value=lambda df: df['value'].astype('float'), date=lambda df: pd.to_datetime(df['date'] ) )
                .set_index('date')
                .rename({'value':seriesID},axis='columns'))
            codeAddendum = f'''\n
                df_output = (df_output[['date','value']]
                    .assign( dropRow = lambda df: pd.to_numeric(df['value'],errors='coerce')   )
                    .dropna()
                    .drop('dropRow',axis=1)
                    .assign(value=lambda df: df['value'].astype('float'), date=lambda df: pd.to_datetime(df['date'] ) )
                    .set_index('date')
                    .rename({{'value': '{seriesID}' }},axis='columns'))
                '''
            cleanCode += dedent(codeAddendum)
            #TODO: relabel value column with symbol
        
        warnings.filterwarnings("ignore", category=UserWarning)
        setattr(df_output, '_meta', dict(filter(lambda entry: entry[0] != dataKey, retrivedData.json().items())))  
        warnings.filterwarnings("always", category=UserWarning)
        return((df_output,cleanCode))
    
    def cleanOutputGeo(self, dbName, dbParams,query, retrivedData): #categories, releases, sources, tags
        if dbName == 'shapes':
            dataKey = query['params_dict']['shape']
        elif dbName == 'series' or dbName == 'data':
            #reproducible code
            cleanCode =  "includeDate = lambda key, array: [ dict(**entry,**{'_date':key}) for entry in array   ]"
            cleanCode += "\ndictData = [  includeDate(key,array) for key,array in  retrivedData.json()['meta']['data'].items() ]"
            cleanCode += "\ndictDataFlat = [item for sublist in dictData for item in sublist]"
            cleanCode += "\ndf_output = pd.DataFrame( dictDataFlat )"
            
            #create dataframe
            includeDate = lambda key, array: [ dict(**entry,**{'_date':key}) for entry in array   ]
            dictData = [  includeDate(key,array) for key,array in  retrivedData.json()['meta']['data'].items() ]
            dictDataFlat = [item for sublist in dictData for item in sublist]
            df_output = pd.DataFrame( dictDataFlat )
            
            #dataframe metadata
            jsonMeta = retrivedData.json()['meta']
            warnings.filterwarnings("ignore", category=UserWarning)
            setattr(df_output, '_meta', dict(filter(lambda entry: entry[0] != 'data', jsonMeta.items())))  
            warnings.filterwarnings("always", category=UserWarning)
            
            return((df_output,cleanCode))
        else: 
            dataKey = dbParams[dbName]['json key']
        
        cleanCode = "df_output =  pd.DataFrame( retrivedData.json()['{}'] )".format(dataKey)
        df_output = pd.DataFrame(retrivedData.json()[dataKey])  # TODO: deal with xml
        warnings.filterwarnings("ignore", category=UserWarning)
        setattr(df_output, '_meta', dict(filter(lambda entry: entry[0] != dataKey, retrivedData.json().items())))  
        warnings.filterwarnings("always", category=UserWarning)
        return((df_output,cleanCode))

class transformIncludeCodeSnippet():
    def transformIncludeCodeSnippet(self,query,baseRequest,userSettings={},pandasCode=""):      
        #load code header - get keys
        apiCode = self.getApiCode(query,userSettings)
        
        #load request's code
        queryCode = self.getQueryCode(query,baseRequest,pandasCode)
        
        return(apiCode + queryCode)
    
    def getQueryCode(self,query,baseRequest,pandasCode=""):
        queryClean = {'url':query['url'],'params':query['params']}  #passing only these two entries of query; params_dict is dropped.
        queryClean['url'] = 'url'
        queryClean['params']=queryClean['params'].replace(baseRequest['params']['api_key'],'{}')+'.format(key)'  #replace explicit api key by the var "key" poiting to it.
        
        queryCode = '''\
            query = {}
            retrivedData = requests.get(**query)
            
            {} #replace json by xml if this is the request format
        '''
        
        queryCode = dedent(queryCode).format(json.dumps(queryClean),pandasCode)
        queryCode = queryCode.replace('"url": "url"', '"url": url')
        queryCode = queryCode.replace('.format(key)"', '".format(key)')
        queryCode = queryCode.replace('"UserID": "key"', '"UserID": key')  #TODO: need to handle generic case, UserID, api_key...        
        return(queryCode)

    def getApiCode(self,query,userSettings): 
        r'''
          The base format of a code that can be used to replicate a driver using Requests directly.
        '''
        try:
            url = query['url']
            if userSettings:  
                apiKeyPath  = userSettings['ApiKeysPath']
                apiKeyLabel = userSettings["ApiKeyLabel"]
            else:
                userSettings = generalSettings.getGeneralSettings( ).userSettings['ApiKeysPath']
                apiKeyPath   = userSettings['ApiKeysPath']
                apiKeyLabel  = userSettings["ApiKeyLabel"]
        except:
            url        = " incomplete connection information "
            apiKeyPath = " incomplete connection information "
        #userSettings = utils.getUserSettings()
        #pkgConfig    = utils.getPkgConfig()
        storagePref  = apiKeyPath.split('.')[-1]
        
        passToCode = {'ApiKeyLabel': apiKeyLabel, "url":url, 'ApiKeysPath':apiKeyPath} #userSettings["ApiKeyLabel"]
        
        code = self.apiCodeOptions(storagePref)
        code = code.format(**passToCode)   
        
        return(code)
    
    def apiCodeOptions(self,storagePref):
        r''''
          storagePref: yaml, json, env
        '''
        if storagePref == 'yaml':
            code = '''\
                import requests
                import yaml 
                import pandas as pd
                
                apiKeysFile = '{ApiKeysPath}'
                with open(apiKeysFile, 'r') as stream:
                    apiInfo= yaml.safe_load(stream)
                    url,key = apiInfo['{ApiKeyLabel}']['url'], apiInfo['{ApiKeyLabel}']['key']
            '''
        elif storagePref == 'json':
            code = '''\
                import requests
                import json    
                import pandas as pd
                
                # json file should contain: {"BEA":{"key":"YOUR KEY","url": "{url}" }
                
                apiKeysFile = '{ApiKeysPath}'
                with open(apiKeysFile) as jsonFile:
                   apiInfo = json.load(jsonFile)
                   url,key = apiInfo['{ApiKeyLabel}']['url'], apiInfo['{ApiKeyLabel}']['key']    
            '''
        else: #default to env
            code = '''\
                import requests
                import os 
                import pandas as pd
                
                url = "{url}"
                key = os.getenv("{ApiKeyLabel}") 
            '''
        return(dedent(code))
    
    def clipcode(self):
        r'''
           Copy the string to the user's clipboard (windows only)
        '''
        try:
            pyperclip.copy(self._lastLoad['code'])
        except:
            print("Loaded session does not have a code entry.  Re-run with verbose option set to True. eg: v.drivername(...,verbose=True)")


class driverMetadata():
    def __call__(self,dbGroup):
        if dbGroup == 'Categories':
            self.metadata = [{
                "displayName": "tags",
                # Name of driver main function - run with getattr(data,'datasetlist')()
                "method": "tags",
                "params": {'file_type': 'json', 'realtime_start': '', 'realtime_end':   '', 'tag_names': '', 'exclude_tag_names': '', 'tag_group_id': '', 'search_text': '', 'limit': '', 'offset': '', 'order_by': '', 'sort_order': ''},
            }]
        elif dbGroup == 'Releases':
            self.metadata = [{
                "displayName": "tags",
                # Name of driver main function - run with getattr(data,'datasetlist')()
                "method": "tags",
                "params": {'file_type': 'json', 'realtime_start': '', 'realtime_end':   '', 'tag_names': '', 'exclude_tag_names': '', 'tag_group_id': '', 'search_text': '', 'limit': '', 'offset': '', 'order_by': '', 'sort_order': ''},
            }]
        elif dbGroup == 'Series':
            self.metadata = [{
                "displayName": "tags",
                # Name of driver main function - run with getattr(data,'datasetlist')()
                "method": "tags",
                "params": {'file_type': 'json', 'realtime_start': '', 'realtime_end':   '', 'tag_names': '', 'exclude_tag_names': '', 'tag_group_id': '', 'search_text': '', 'limit': '', 'offset': '', 'order_by': '', 'sort_order': ''},
            }]
        elif dbGroup == 'Sources':
            self.metadata = [{
                "displayName": "tags",
                # Name of driver main function - run with getattr(data,'datasetlist')()
                "method": "tags",
                "params": {'file_type': 'json', 'realtime_start': '', 'realtime_end':   '', 'tag_names': '', 'exclude_tag_names': '', 'tag_group_id': '', 'search_text': '', 'limit': '', 'offset': '', 'order_by': '', 'sort_order': ''},
            }]
        elif dbGroup == 'Tags':
            self.metadata = [{
                "displayName": "tags",
                # Name of driver main function - run with getattr(data,'datasetlist')()
                "method": "tags",
                "params": {'file_type': 'json', 'realtime_start': '', 'realtime_end':   '', 'tag_names': '', 'exclude_tag_names': '', 'tag_group_id': '', 'search_text': '', 'limit': '', 'offset': '', 'order_by': '', 'sort_order': ''},
            }]
        else:
            self.metadata = [{
                "displayName": "datasetlist",
                # Name of driver main function - run with getattr(data,'datasetlist')()
                "method": "datasetlist",
                "params": {'file_type': 'json', 'realtime_start': '', 'realtime_end':   '', 'tag_names': '', 'exclude_tag_names': '', 'tag_group_id': '', 'search_text': '', 'limit': '', 'offset': '', 'order_by': '', 'sort_order': ''},
            }]
        
        return(self.metadata)



if __name__ == '__main__':
    case = driverCore(dbGroupName = 'Series')
    print(case('gdp',verbose=True))


# content of conftest.py
import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--cmdopt", action="store", default="", help="enter API key"
    )


@pytest.fixture
def cmdopt(request):
    return request.config.getoption("--cmdopt")


import subprocess
import os
from datapungi_fed.utils import getUserSettings


def runTests(outputPath='',testsPath='',verbose = True):
    if not testsPath:
       testsPath =  os.path.dirname(os.path.abspath(__file__)).replace("\\","/")
       print('**************************** \nWill run tests in: ' + testsPath)
    if not outputPath:
        outputPath = "U:/"
        try:
            settingsFile = getUserSettings()
            outputPath = settingsFile['TestsOutputPath']
        except:
            print("Could not load TestOutputPath from user settings.  Perhaps run util.setTestFolder( FilePath )  ")
    subprocess.Popen('pytest ' + testsPath + ' --html='+outputPath+'datapungi_fed_Tests.html --self-contained-html')
    if verbose:
        print('Tests will be saved in '+outputPath+'datapungi_fed_Tests.html \n****************************')

if __name__ == '__main__':
    from sys import argv    
    import subprocess
    import os 

    runTests()
    #print(os.path.dirname(os.path.realpath(__file__)))
    #query = subprocess.Popen('pytest --html=datapungibea_Tests.html')
    #print(query)


"""Gets FRED and GeoFRED data from Federal Reserve (FED) by connecting to its API."""
import pandas
import requests
import sys

from datapungi_fed.api import *
import datapungi_fed.tests as tests

__version__ = '0.2.1'


class topCall(sys.modules[__name__].__class__):
    def __call__(self,*args,**kwargs):
        coreClass = data()
        return(coreClass(*args,**kwargs))
    def __str__(self):
        starter = "\nSample starter: \n\nimport datapungi_fed as dpf \n\ndata = dpf.data() \nprint(data) \npor just query a time series: \ndpf('gdp')"
        return(starter)

sys.modules[__name__].__class__ = topCall